# -*- coding: utf-8 -*-
# Copyright © 2021 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from __future__ import print_function

import os
import platform
import sys

from contrast import __version__
from contrast.agent.patch_controller import disable_assess_patches
from contrast.agent import patch_controller, service_client, scope
from contrast.agent.policy.loader import Policy
from contrast.agent.assess.policy import trigger_policy
from contrast.agent.assess.preflight import update_preflight_hashes
from contrast.agent.assess.rules.providers.enable import enable_providers
from contrast.agent.middlewares.route_coverage.routes_mixin import RoutesMixin
from contrast.agent.settings_state import SettingsState
from contrast.agent.speedracer_input_analysis import get_input_analysis
from contrast.api.dtm_pb2 import AttackResult
from contrast.utils.decorators import cached_property
from contrast.utils.exceptions.contrast_service_exception import (
    ContrastServiceException,
)
from contrast.extern.six import PY2
from contrast.utils.exceptions.security_exception import SecurityException
from contrast.utils.library_reader.library_reader import LibraryReader
from contrast.utils.loggers.logger import setup_agent_logger, setup_basic_agent_logger

# initialize a basic logger until config is parsed
logger = setup_basic_agent_logger()


class BaseMiddleware(RoutesMixin):
    """
    BaseMiddleware contains all the initial setup for the framework middlewares

    Requirements:

        1. It's callable
        2. It has call_with_agent
        3. It has call_without_agent

    Pre and post filter calls should not block the flow that this class has.

    Pre -> get_response -> post
    """

    _loaded = False
    app_name = ""  # This should be overridden by child classes

    DIAGNOSTIC_ENDPOINT = "/save-contrast-security-config"
    DIAGNOSTIC_ALLOWED_SERVER = "localhost"
    DIAGNOSTIC_ALLOWED_IP = "127.0.0.1"
    OVERRIDE_MESSAGE = (
        "A security filter prevented original response from being returned."
    )
    LOGS_SEPARATOR = "-" * 120
    PY2_DEPRECATION_MSG = (
        "[Contrast Security] [DEPRECATION] Support for Python 2.7  will be removed in "
        "2021. Please contact Customer Support prior if you require continued support."
    )

    def __init__(self):
        # use _loaded to call our middleware only once because of runserver and
        # WSGIHandler being initialized twice.
        if BaseMiddleware._loaded:
            return

        # id to distinguish middlewares created by different processes
        self.id = id(self)

        self.log_deprecation()
        self.log_initialize()

        self.settings = None

        self.routes = {}
        self.first_request = True

        if not self.initialize_settings(self.app_name):
            logger.error("Unable to initialize Contrast Agent.")
            return

        if not self.settings.is_agent_config_enabled():
            logger.warning("Contrast Agent is not enabled.")
            return

        setup_agent_logger(self.settings.config)
        self.settings.config.log_config()

        if not self.initialize_service_client():
            logger.error("Unable to initialize Contrast Agent.")
            return

        # This call must happen after initializing SR so we have all data to log
        self.log_environment()
        self.warn_for_misleading_config()

        self.initialize_libraries()

        patch_controller.enable_patches()

        if self.settings.is_assess_enabled():
            enable_providers()
            self.scan_configs()

        self.log_finish_initialize()

        BaseMiddleware._loaded = True

    def initialize_settings(self, app_name):
        """
        Initialize agent settings.

        Returns True on settings being initialized and False if any failure.
        """
        try:
            self.settings = SettingsState(app_name=app_name)
        except Exception as e:
            logger.error(e)
            return False
        return True

    def initialize_service_client(self):
        """
        Initialize connection to service.

        Returns True on service being initialized and False if any failure.
        """
        try:
            service_client.send_startup_messages()
        except (Exception, ContrastServiceException) as e:
            logger.error(e)
            return False
        return True

    def warn_for_misleading_config(self):
        protect_enabled = self.settings.is_protect_enabled()
        assess_enabled = self.settings.is_assess_enabled()

        logger.info("Protect: %s", protect_enabled)
        logger.info("Assess: %s", assess_enabled)

        if protect_enabled and self.settings.config.get("assess.enable"):
            logger.warning("Protect is running but Assess is enabled in local config")
            logger.warning("Defaulting to Protect behavior only")

        if not protect_enabled and self.settings.config.get("protect.enable", True):
            logger.warning("Protect enabled in local config but disabled by Teamserver")

        if (
            self.settings.is_agent_config_enabled()
            and not protect_enabled
            and not assess_enabled
        ):
            logger.warning("Neither Protect nor Assess is running")

    def log_environment(self):
        """
        Log current working directory, python version and pip version
        """
        banner = "{0}ENVIRONMENT{0}".format("-" * 50)
        logger.debug(banner)
        logger.debug("Current Working Dir: %s", os.getcwd())
        logger.debug("Python Version: %s", sys.version)
        logger.debug("Framework Version: %s", self.settings.framework)
        logger.debug("Contrast Python Agent Version: %s", __version__)
        logger.debug(
            "Contrast Service Version %s",
            self.settings.server_features.contrast_service,
        )
        logger.debug("Platform %s", platform.platform())

        try:
            import pip

            logger.debug("Pip Version: %s", pip.__version__)
        except:
            pass

        logger.debug(banner)

    def initialize_libraries(self):
        """
        If enabled, read libraries from the application
        :return: True
        """
        if not self.settings.is_analyze_libs_enabled():
            return

        # Passing callback to send message due to a circular import issue, and a
        # deadlock occurring if something is imported inside of a function running in
        # this thread
        self.library_reader = LibraryReader(
            send_message_func=service_client.send_messages
        )
        self.library_reader.start_library_analysis_thread()

    def is_agent_enabled(self):
        """
        Agent is considered enabled if all of the following are true:
        1. config value for 'enable' is True (or empty, defaults to True)
            (meaning no errors during initialization process including initial
            connection to Speedracer)
        2. ServiceClient connection to Speedracer is True

        NOTE: If #3 is false (the connection to Speedracer is down at any time during
        the request cycle) then the agent is automatically disabled.
        """
        if self.settings is None:
            return False

        if not self.settings.is_agent_config_enabled():
            return False

        if not service_client.is_connected():
            try:
                service_client.send_startup_messages()
            except ContrastServiceException:
                return False

        return service_client.is_connected()

    def call_with_agent(self, context, environ, start_response):
        pass

    def call_without_agent(self):
        """
        If Agent is either disabled or not initialized due to either entire settings
        initialization failing or just loggers failing, then we should print to stdout
        that the Agent is not running.

        In the case of the agent is disabled by a reaction from SR, we have to disable
        assess patches here because checking for context inside of our builtin string
        patches is much more expensive to do since they are called very frequently.

        The agent does not set context when this function is called so all other patches
        (e.g propagators) that check context shouldn't run.
        """
        disable_assess_patches()

        logger.warning("Contrast Agent is not running.")

    def handle_ensure(self, context, request):
        """
        Method that should run for all middlewares AFTER every request is made.
        """
        if context is None:
            logger.error("Context not defined in middleware ensure")
            return

        with scope.contrast_scope():
            if self.settings.is_assess_enabled():
                # route discovery and storage needs to occur before sending messages
                self.handle_routes(context, request)

                logger.debug("Updating preflight hashes with route info")
                update_preflight_hashes(context)

            logger.debug("Sending final messages for reporting.")

            service_client.send_messages(self.final_messages(context))

    def final_messages(self, context):
        final_messages = [
            context.server_activity,
            context.activity,
        ]

        # Currently we do not report an observed route if the route signature is empty.
        # As a team we've decided there isn't a meaningful default signature value
        # we can provide to customers. If a route doesn't show up in Contrast UI,
        # it may be due to its missing signature. In this scenario, we will have to work
        # with the customer directly to understand why the signature was not created.
        if self.settings.is_assess_enabled() and context.observed_route.signature:
            final_messages.append(context.observed_route)

        if self.first_request:
            # ApplicationUpdate message only sent at first request once we have routes
            final_messages.append(service_client.build_update_message(self.routes))
            self.first_request = False

        return final_messages

    def generate_security_exception_response(self):
        """
        Generate a response to be returned in the case of a SecurityException.

        Middlewares must override this method in order to provide the kind
        of response object that is expected by that framework.
        """
        raise NotImplementedError(
            "Middlewares must provide their own security response generator"
        )

    def handle_exception(self, exception):
        """
        Handle an exception being thrown especially if it is a SecurityException
        """
        logger.debug(
            "Handling %s raised in %s",
            exception.__class__.__name__,
            self.__class__.__name__,
        )
        if isinstance(exception, SecurityException):
            logger.info("%s: %s", exception.__class__.__name__, exception)
            logger.debug("Overriding response in Contrast Middleware")
            return self.generate_security_exception_response()

        logger.error("Reraising %r", exception)
        raise

    def prefilter(self, context):
        """
        Prefilter - AKA input analysis - is performed mostly in Speedracer but partly
        in the agent.

        This is PROTECT only.

        In this method we call on speedracer to do input analysis, which can result in:
        1. Speedracer finds an attack in which case we block the request
        2. Speedracer returns input analysis to use for later sink / infilter analysis,
            in which case we store it here in the request context.
        """
        if not self.settings.is_protect_enabled():
            return

        context.timer.set_start("prefilter")

        logger.debug("Running prefilter.")
        context.speedracer_input_analysis = get_input_analysis()

        self.agent_prefilter()

        logger.debug("Completed prefilter.")

        context.timer.set_end("prefilter")

    def agent_prefilter(self):
        """
        Prefilter for any rules that do not yet use speedracer
        """
        scope.enter_contrast_scope()
        try:
            self.prefilter_defend()
        except SecurityException:
            logger.exception("PROTECT: threw security exception in prefilter")
            raise
        except:
            logger.exception("PROTECT: threw exception in prefilter")
        finally:
            scope.exit_contrast_scope()

    def prefilter_defend(self):
        rules = self.settings.defend_rules
        logger.debug("PROTECT: Running Agent prefilter.")

        for rule in rules.values():
            is_prefilter = rule.is_prefilter()

            if is_prefilter:
                rule.prefilter()

    def postfilter(self, context):
        """
        For all postfilter enabled rules.
        """
        context.timer.set_start("postfilter")

        if self.settings.is_protect_enabled():
            try:
                self.postfilter_defend()
            except SecurityException as e:
                logger.exception("PROTECT: threw security exception in postfilter")
                raise e
            except Exception:
                logger.exception("PROTECT: threw exception in postfilter")

        if self.settings.is_assess_enabled():
            scope.enter_contrast_scope()
            try:
                self.postfilter_assess(context)
            except Exception:
                logger.exception("ASSESS: threw exception in postfilter")
            scope.exit_contrast_scope()

        context.timer.set_end("postfilter")

    def _process_trigger_handler(self, handler):
        """
        Gather metadata about response handler callback for xss trigger node

        We need to check whether the response handler callback is an instance method or
        not. This affects the way that our policy machinery works, and it also affects
        reporting, so we need to make sure to account for the possibility that handler
        is a method of some class rather than a standalone function.

        This should be called by the `trigger_node` method in child classes.
        """
        module = handler.__module__
        class_name = ""

        if hasattr(handler, "__self__"):
            class_name = handler.__self__.__class__.__name__
            args = (handler.__self__,)
            instance_method = True
        else:
            args = ()
            instance_method = False

        return module, class_name, args, instance_method

    @cached_property
    def trigger_node(self):
        """
        Trigger node property used by assess reflected xss postfilter rule

        This must be overridden by child classes that make use of the reflected
        xss postfilter rule.
        """
        raise NotImplementedError("Children must define trigger_node property")

    def assess_reflected_xss_postfilter(self, context):
        """
        Evaluate xss rule for assess

        The xss rule is applied to the response body in order to determine whether it
        contains untrusted data. We rely on propagation through any template rendering
        and through the framework. The expectation is that untrusted data from
        a request will propagate successfully all the way to the response body, which
        we are able to see here.

        Each child middleware class must implement specific logic for the trigger node
        since the reporting will differ between frameworks.
        """
        policy = Policy()

        rule = policy.triggers["reflected-xss"]

        # We need to exit scope here in order to account for the fact that some
        # frameworks evaluate the content lazily. We don't want to be in scope when
        # that occurs since it would make us lose propagation. This would prevent us
        # from seeing the response as a tracked string, which we require in order to
        # apply the rule.
        with scope.pop_contrast_scope():
            result = context.response.body

        trigger_node, args = self.trigger_node

        # TODO: PYT-721 this trigger scope is already and shouldn't be required.
        # If/when we consolidate all scope levels, this scope should be removed
        # entirely since the postfilter is already in scope when it's called.
        with scope.trigger_scope():
            trigger_policy.apply(rule, [trigger_node], result, args, {})

    def postfilter_assess(self, context):
        """
        Run postfilter for any assess rules. Reflected xss rule runs by default.
        May be overridden in child classes.

        If the response content type matches a allowed content type, do not run
        xss postfilter assess This is because the security team
        considers reflected xss within these content types to be a false positive.
        """
        logger.debug("ASSESS: Running Agent postfilter.")

        # it is possible other forms (capital case) variants exist
        accepted_xss_response_content_types = [
            "/csv",
            "/javascript",
            "/json",
            "/pdf",
            "/x-javascript",
            "/x-json",
            "/plain",
        ]

        content_type = context.response.headers.get("content-type", "")

        if not any(
            [
                name
                for name in accepted_xss_response_content_types
                if name in content_type
            ]
        ):
            self.assess_reflected_xss_postfilter(context)

    def postfilter_defend(self):
        rules = self.settings.defend_rules
        logger.debug("PROTECT: Running Agent postfilter.")

        for rule in rules.values():
            if rule.is_postfilter():
                rule.postfilter()

    def check_for_blocked(self, context):
        """
        Checks for BLOCK events in case SecurityException was caught by app code

        This should be called by each middleware after the view is generated
        but before returning the response (it can be before or after
        postfilter).

        If we make it to this call, it implies that either no SecurityException
        occurred, or if one did occur, it was caught by the application. If we
        find a BLOCK here, it necessarily implies that an attack was detected
        in the application, but the application caught our exception. If the
        application hadn't caught our exception, we never would have made it
        this far because the exception would have already bubbled up to the
        middleware exception handler. So this is really our first and our last
        opportunity to check for this particular edge case.
        """
        for result in context.activity.results:
            if result.response == AttackResult.BLOCKED:
                msg = "Rule {} threw a security exception".format(result.rule_id)
                raise SecurityException(None, message=msg)

    def log_start_request_analysis(self, request_path):
        logger.debug("Beginning request analysis", request_path=request_path)

    def log_end_request_analysis(self, request_path):
        logger.debug(
            "Ending request analysis", request_path=request_path,
        )

    def log_deprecation(self):
        if PY2:
            warn_color = "\033[91m"
            end_color = "\033[0m"
            print("\n{}{}{}\n".format(warn_color, self.PY2_DEPRECATION_MSG, end_color))
            logger.warning(self.PY2_DEPRECATION_MSG)

    def log_initialize(self):
        logger.info(
            "Initializing Contrast Agent %s [id=%s]", self.__class__.__name__, self.id
        )

        logger.info("Contrast Python Agent Version: %s\n", __version__)

    def log_finish_initialize(self):
        logger.info(
            "Finished Initializing Contrast Agent %s [id=%s] \n\n%s\n",
            self.__class__.__name__,
            self.id,
            self.LOGS_SEPARATOR,
        )

    def scan_configs(self):
        """
        Run config scanning rules for assess

        Not all frameworks we support will necessarily have config scanning rules.
        Those that do should override this method.
        """
        logger.debug("No config scanning rules for %s", self.__class__.__name__)



# -*- coding: utf-8 -*-
# Copyright © 2021 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from collections import OrderedDict

from contrast.agent.protect.rule.cmdi_rule import CmdInjection
from contrast.agent.protect.rule.deserialization_rule import Deserialization
from contrast.agent.protect.rule.http_method_tampering import MethodTampering
from contrast.agent.protect.rule.malformed_header import MalformedHeader
from contrast.agent.protect.rule.nosqli_rule import NoSqlInjection
from contrast.agent.protect.rule.path_traversal_rule import PathTraversal
from contrast.agent.protect.rule.sqli_rule import SqlInjection
from contrast.agent.protect.rule.ssrf_rule import Ssrf
from contrast.agent.protect.rule.unsafe_file_upload_rule import UnsafeFileUpload
from contrast.agent.protect.rule.xss_rule import Xss
from contrast.agent.protect.rule.xxe_rule import Xxe


class RulesBuilder(object):
    """
    Combines all rules
    """

    @staticmethod
    def build_protect_rules(settings):
        """
        Build a dict with rules with prefilter rules first.
        We want prefilter rules first so they get evaluated / trigger first.
        :param settings:
        :return: an ordered dict of protect rules
        """
        rules = OrderedDict(
            {
                UnsafeFileUpload.NAME: UnsafeFileUpload(settings),
                CmdInjection.NAME: CmdInjection(settings),
                Deserialization.NAME: Deserialization(settings),
                # Turned off until TS can handle rule information
                MalformedHeader.NAME: MalformedHeader(settings),
                MethodTampering.NAME: MethodTampering(settings),
                NoSqlInjection.NAME: NoSqlInjection(settings),
                # Padding Oracle rule is currently disabled - CONTRAST-35352
                # PaddingOracle.NAME: PaddingOracle(settings),
                PathTraversal.NAME: PathTraversal(settings),
                SqlInjection.NAME: SqlInjection(settings),
                Ssrf.NAME: Ssrf(settings),
                Xss.NAME: Xss(settings),
                Xxe.NAME: Xxe(settings),
            }
        )

        return rules



# -*- coding: utf-8 -*-
# Copyright © 2021 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.extern.wrapt import register_post_import_hook

import contrast
from contrast.patches.databases import dbapi2
from contrast.utils import inventory_utils
from contrast.utils.patch_utils import patch_cls_or_instance

from contrast.extern import structlog as logging

logger = logging.getLogger("contrast")

PYMYSQL = "pymysql"
VENDOR = "MySQL"


def patch_connect(module):
    def _connect(orig_func, patch_policy=None, *args, **kwargs):
        """Record DB inventory for MySQL"""
        try:
            context = contrast.CS__CONTEXT_TRACKER.current()
            if context is not None:
                db_inventory = dict(
                    vendor=VENDOR,
                    host=kwargs.get("host"),
                    port=kwargs.get("port"),
                    database=kwargs.get("db"),
                )
                inventory_utils.append_db(context.activity, db_inventory)
        except Exception:
            logger.exception("Failed to add inventory for %s", VENDOR)

        return orig_func(*args, **kwargs)

    patch_cls_or_instance(module, "connect", _connect)


def instrument_pymysql(pymysql):
    """
    Note: Unlike other sql drivers, pymysql does not have an executescript method.
    """
    patch_connect(pymysql)
    dbapi2.instrument_cursor(PYMYSQL, pymysql.cursors.Cursor)


def register_patches():
    register_post_import_hook(instrument_pymysql, PYMYSQL)



# -*- coding: utf-8 -*-
# Copyright © 2021 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.disable_reaction import DisableReaction
from contrast.api.settings_pb2 import Reaction

from contrast.extern import structlog as logging

logger = logging.getLogger("contrast")


class ReactionProcessor(object):
    @staticmethod
    def process(application_settings, settings):
        if (
            application_settings is None
            or settings is None
            or application_settings.reactions is None
            or len(application_settings.reactions) == 0
        ):
            return

        for reaction in application_settings.reactions:
            logger.debug("Received the following reaction: %s", reaction.operation)

            if reaction.operation == Reaction.DISABLE:
                DisableReaction.run(settings)



# -*- coding: utf-8 -*-
# Copyright © 2021 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.middlewares.route_coverage.coverage_utils import (
    CoverageUtils,
    DEFAULT_ROUTE_METHODS,
)

DEFAULT_ROUTE_METHODS = DEFAULT_ROUTE_METHODS + ("PUT", "PATCH", "DELETE")


def create_bottle_routes(app):
    """
       Returns all the routes registered to a Bottle app as a dict
       :param app: Bottle app
       :return: dict {route_id:  RouteCoverage}
       """
    routes = {}
    for rule in app.routes:
        view_func = rule.callback
        route = build_bottle_route(rule.rule, view_func)
        route_id = str(id(view_func))

        for method_type in DEFAULT_ROUTE_METHODS:
            key = CoverageUtils.build_key(route_id, method_type)
            routes[key] = CoverageUtils.build_route_coverage(
                verb=method_type,
                url=CoverageUtils.get_normalized_uri(str(rule.rule)),
                route=route,
            )
    return routes


def build_bottle_route(view_func_name, view_func):
    view_func_args = CoverageUtils.build_args_from_function(view_func)
    return view_func_name + view_func_args



# -*- coding: utf-8 -*-
# Copyright © 2021 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from .cs_str import set_attr_on_type


def smart_setattr(owner, name, patch):
    """
    Use either setattr or set_attr_on_type as appropriate
    """
    # For some reason, set_attr_on_type doesn't work for special methods
    # We may eventually need to refine this logic
    force_patch = isinstance(owner, type) and not (
        name.startswith("__") and name.endswith("__")
    )

    setattr(owner, name, patch) if not force_patch else set_attr_on_type(
        owner, name, patch
    )



# -*- coding: utf-8 -*-
# Copyright © 2021 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import pytest

from contrast.agent.middlewares.base_middleware import BaseMiddleware
from contrast.api.dtm_pb2 import UserInput
from contrast.api.settings_pb2 import (
    AgentSettings,
    InputAnalysis,
    InputAnalysisResult,
    ProtectState,
)
from contrast.extern import six


class _MockedInputAnalysisMixin(object):
    def mock_input_analysis(self, mocker, rule_id, attack, security_exception=False):
        """
        Assemble a fake response for speedracer's input analysis, then mock
        communication with speedracer during input analysis to return this response.
        """
        fake_state = ProtectState(
            uuid="012345678987654",
            track_request=True,
            security_exception=security_exception,
        )
        fake_result = InputAnalysisResult(
            rule_id=rule_id,
            value=attack,
            path="user_input",
            key="user_input",
            input_type=UserInput.PARAMETER_VALUE,
            score_level=InputAnalysisResult.DEFINITEATTACK,
            ids=[],
        )
        fake_analysis = InputAnalysis(results=[fake_result])
        fake_response = AgentSettings(
            protect_state=fake_state, input_analysis=fake_analysis
        )
        mocker.patch(
            "contrast.agent.speedracer_input_analysis.service_client.send_messages_get_responses",
            return_value=[fake_response],
        )

    @pytest.mark.parametrize(
        "trigger", ["sqlite3-execute", "sqlite3-executemany", "sqlite3-executescript"]
    )
    def test_sqli(self, trigger, mocker):
        rule_id = "sql-injection"
        attack = "foo' OR 1=1; DROP TABLE Users--"
        self.mock_input_analysis(mocker, rule_id, attack)

        response = self.client.get(
            "/vulnpy/sqli/{}".format(trigger),
            {"user_input": attack},
            status=403,
            expect_errors=True,
        )

        # 3 queries per vulnpy request: database reset, attack, fetch results
        # we skip the last one here because the second causes a SecurityException
        assert self.request_context.activity.query_count == 2
        self.assert_blocked(response, rule_id)

    @pytest.mark.parametrize("trigger", ["io-open", "open"])
    def test_path_traversal(self, trigger, mocker):
        rule_id = "path-traversal"
        attack = "../../../../etc/passwd"
        self.mock_input_analysis(mocker, rule_id, attack)

        response = self.client.get(
            "/vulnpy/pt/{}".format(trigger),
            {"user_input": attack},
            status=403,
            expect_errors=True,
        )

        self.assert_blocked(response, rule_id)

    def test_xss(self, mocker):
        """
        XSS is BAP only - we need SR to tell us to raise a security exception
        """
        rule_id = "reflected-xss"
        attack = "<script>alert(1);</script>"
        self.mock_input_analysis(mocker, rule_id, attack, security_exception=True)

        response = self.client.get(
            "/vulnpy/xss/raw", {"user_input": attack}, status=403, expect_errors=True
        )

        self.assert_blocked(response, rule_id)

    @pytest.mark.parametrize(
        "trigger",
        [
            "legacy-urlopen",
            "urlopen-str",
            "httpconnection-putrequest-method",
            "httpsconnection-putrequest-method",
            "httpconnection-request-method",
            "httpsconnection-request-method",
        ],
    )
    def test_ssrf(self, trigger, mocker):
        """
        This attack vector doesn't really make sense with the `-method` triggers.
        However, we wouldn't flag something like BADVERB as an attack, so this is
        probably still the most realistic scenario.
        """
        rule_id = "ssrf"
        attack = "http://www.attacker.com/evil"
        self.mock_input_analysis(mocker, rule_id, attack)

        response = self.client.get(
            "/vulnpy/ssrf/{}".format(trigger),
            {"user_input": attack},
            status=403,
            expect_errors=True,
        )

        self.assert_blocked(response, rule_id)

    @pytest.mark.parametrize(
        "trigger",
        [
            "httpconnection-putrequest-url",
            "httpsconnection-putrequest-url",
            "httpconnection-request-url",
            "httpsconnection-request-url",
        ],
    )
    def test_ssrf_safe(self, trigger, mocker):
        """
        For assess, labs determined that the URL itself (without protocol/host)
        isn't vulnerable to SSRF. We extended this logic to protect.
        """
        attack = "http://www.attacker.com/evil"
        self.mock_input_analysis(mocker, "ssrf", attack)

        self.client.get(
            "/vulnpy/ssrf/{}".format(trigger), {"user_input": attack}, status=200,
        )

    def test_unsafe_file_upload(self, mocker):
        """
        This test doesn't prove much, for now. It's more of a placeholder for when
        we start using a real connection to the service.
        """
        rule_id = "unsafe-file-upload"
        attack = "some_dangerous_file.php"
        self.mock_input_analysis(mocker, rule_id, attack, security_exception=True)

        response = self.client.post(
            "/vulnpy",
            upload_files=[("user_file", attack, b"file content doesn't matter")],
            status=403,
            expect_errors=True,
        )

        self.assert_blocked(response, rule_id)

    @pytest.mark.parametrize(
        "attack",
        [
            # a word of caution: these execute for real on your machine
            "grep 'passwd' && echo",
            "echo a | boom",
            "/bin/sh -c ls -al; cat /etc/passwd",
        ],
    )
    @pytest.mark.parametrize("trigger", ["subprocess-popen", "os-system"])
    def test_cmdi_command_backdoor_disabled(self, trigger, attack, mocker):
        """
        Here's a really interesting case - which directly tests CMDi REP
        detect_chained_commands. The input analysis is for CMDi, but a totally
        different string. However, we run detect_chained_commands on the
        string that makes it to the trigger as long as there is at least one
        evaluation result for cmd-injection.
        """
        self.settings.config.put(
            "protect.rules.cmd-injection.detect_parameter_command_backdoors", False
        )

        rule_id = "cmd-injection"
        self.mock_input_analysis(mocker, rule_id, "a totally unrelated string")

        response = self.client.get(
            "/vulnpy/cmdi/{}".format(trigger),
            {"user_input": "grep 'passwd' && echo"},
            status=403,
            expect_errors=True,
        )

        self.assert_blocked(response, rule_id)

    @pytest.mark.parametrize("trigger", ["subprocess-popen", "os-system"])
    def test_cmdi_command_backdoor_disabled_safe(self, trigger, mocker):
        """
        Same as above, but without malicious input; leads to 200 instead of 403.
        """
        self.settings.config.put(
            "protect.rules.cmd-injection.detect_parameter_command_backdoors", False
        )

        self.mock_input_analysis(mocker, "cmd-injection", "a totally unrelated string")

        self.client.get(
            "/vulnpy/cmdi/{}".format(trigger),
            {"user_input": "harmless input"},
            status=200,
        )


class _NoInputAnalysisMixin(object):
    @pytest.mark.parametrize("trigger", ["subprocess-popen", "os-system"])
    def test_cmdi_command_backdoor_no_input_analysis(self, trigger):
        """
        When an argument to a CMDi trigger matches an HTTP parameter, the
        agent detects a "command backdoor" and blocks. This occurs even
        without input analysis from speedracer.
        """
        attack = "interestingly this does not matter"
        response = self.client.get(
            "/vulnpy/cmdi/{}".format(trigger),
            {"user_input": attack},
            status=403,
            expect_errors=True,
        )

        self.assert_blocked(response, "cmd-injection")

    @pytest.mark.parametrize("trigger", ["subprocess-popen", "os-system"])
    def test_cmdi_command_backdoor_disabled_no_input_analysis(self, trigger):
        """
        With no input analysis AND command backdoors disabled, even real attacks
        will not lead to a 403 response.
        """
        self.settings.config.put(
            "protect.rules.cmd-injection.detect_parameter_command_backdoors", False
        )

        attack = "grep 'passwd' && echo"
        self.client.get(
            "/vulnpy/cmdi/{}".format(trigger), {"user_input": attack}, status=200
        )

    @pytest.mark.parametrize(
        "trigger", ["yaml-load", "yaml-load-all", "pickle-load", "pickle-loads"]
    )
    @pytest.mark.parametrize(
        "attack",
        [
            '!!map { ? !!str "goodbye" : !!python/object/apply:subprocess.check_output [ !!str "ls", ], }',
            "cos system (S'/bin/sh' tR.",
        ],
    )
    def test_deserialization_no_input_analysis(self, trigger, attack):
        """
        Both deserialization triggers (yaml and pickle) check for attacks that would only work
        against specific triggers. It's possible that in the future we may want to make this
        more specific. However, for now we're free to mix and match attacks and triggers.
        """
        response = self.client.get(
            "/vulnpy/deserialization/{}".format(trigger),
            {"user_input": attack},
            status=403,
            expect_errors=True,
        )

        self.assert_blocked(response, "untrusted-deserialization")

    @pytest.mark.parametrize(
        "trigger",
        ["lxml-etree-fromstring", "xml-dom-pulldom-parsestring", "xml-sax-parsestring"],
    )
    def test_xxe_no_input_analysis(self, trigger):
        """
        This particular attack is designed to work against lxml specifically.
        However, all 3 XXE triggers will still see this as an attack.
        """
        attack = (
            '<!DOCTYPE doc [ <!ENTITY lxml SYSTEM file:///etc/passwd"> ]>\n'
            "<root>\n"
            "<element>&lxml;</element>\n"
            "</root>\n"
        )
        response = self.client.get(
            "/vulnpy/xxe/{}".format(trigger),
            {"user_input": attack},
            status=403,
            expect_errors=True,
        )

        self.assert_blocked(response, "xxe")

    @pytest.mark.parametrize(
        "attack", ["../../../../etc/passwd", "super/duper/../../etc/hosts", "::$DATA"],
    )
    @pytest.mark.parametrize("trigger", ["io-open", "open"])
    def test_path_traversal_no_input_analysis_unsafe(self, trigger, attack):
        response = self.client.get(
            "/vulnpy/pt/{}".format(trigger),
            {"user_input": attack},
            status=403,
            expect_errors=True,
        )

        self.assert_blocked(response, "path-traversal")

    @pytest.mark.parametrize("trigger", ["io-open", "open"])
    def test_path_traversal_no_input_analysis_safe(self, trigger):
        self.client.get(
            "/vulnpy/pt/{}".format(trigger), {"user_input": "normal"}, status=200,
        )


class VulnpyProtectTestMixin(_MockedInputAnalysisMixin, _NoInputAnalysisMixin):
    def assert_blocked(self, response, rule_id):
        assert response.status_int == 403
        assert six.ensure_str(response.body) == BaseMiddleware.OVERRIDE_MESSAGE
        assert len(self.request_context.activity.results) == 1
        assert self.request_context.activity.results[0].rule_id == rule_id



# -*- coding: utf-8 -*-
# Copyright © 2021 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import abc

from contrast.extern import six


@six.add_metaclass(abc.ABCMeta)
class BaseResponseWrapper(object):
    @property
    @abc.abstractmethod
    def body(self):
        """
        Response body as bytes
        """
        pass

    @property
    @abc.abstractmethod
    def headers(self):
        """
        Response headers as a dictionary-like object.

        This object must support normalized lowercase lookup. For example:
        my_response_wrapper.headers['content-length']

        In order to properly extract response headers for analysis, we currently expect
        this field to be a multidict that implements a method called `dict_of_lists()`.
        """
        pass

    @property
    @abc.abstractmethod
    def status_code(self):
        """
        Status code as an integer
        """
        pass



# -*- coding: utf-8 -*-
# Copyright © 2021 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.extern.wrapt import register_post_import_hook
from contrast.patches.pycrypto import cs__pycrypto_decrypt, DECRYPT
from contrast.utils.patch_utils import patch_cls_or_instance


def patch_pycryptodomex(cryptodome_module):
    patch_cls_or_instance(
        cryptodome_module.Cipher._mode_cbc.CbcMode, DECRYPT, cs__pycrypto_decrypt
    )


def register_patches():
    register_post_import_hook(patch_pycryptodomex, "Cryptodome")



# -*- coding: utf-8 -*-
# Copyright © 2021 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import sys

from contrast.extern.wrapt import register_post_import_hook

import contrast
from contrast.agent import scope
from contrast.agent.policy import patch_manager
from contrast.agent.settings_state import SettingsState
from contrast.agent.assess.apply_trigger import cs__apply_trigger
from contrast.agent.policy.trigger_node import TriggerNode
from contrast.agent.policy.loader import Policy
from contrast.utils.decorators import fail_safely
from contrast.utils.patch_utils import patch_cls_or_instance

from contrast.extern import structlog as logging

logger = logging.getLogger("contrast")


@fail_safely("Failed to apply rule for crypto-bad-mac")
def apply_rule(args, kwargs):
    context = contrast.CS__CONTEXT_TRACKER.current()

    if context is None or not SettingsState().is_assess_enabled():
        return

    if scope.in_contrast_scope():
        return

    with scope.contrast_scope():
        policy = Policy()

        bad_mac_rule = policy.triggers["crypto-bad-mac"]

        self_obj = args[0]

        trigger_node = TriggerNode(
            self_obj.__class__.__module__,
            self_obj.__class__.__name__,
            True,
            "__init__",
            None,
            dataflow=False,
        )

        cs__apply_trigger(
            context,
            bad_mac_rule,
            trigger_node,
            None,
            self_obj,
            None,
            None,
            args,
            kwargs,
        )


def __init__(orig_func, patch_policy=None, *args, **kwargs):
    orig_func(*args, **kwargs)
    # TODO: PYT-798 This may be currently broken because there are not policy nodes associated
    # with these patch locations. These need to be added and this rule needs to be tested.
    apply_rule(args, kwargs)


def build_hash_patch(name):
    def patch_hash(module):
        hasher = getattr(module, name, None)
        if hasher is None:
            logger.debug(
                "WARNING: failed to instrument %s.%s. "
                "Unsupported pycryptodome version likely",
                module.__name__,
                name,
            )
            return

        patch_cls_or_instance(hasher, "__init__", __init__)

    return patch_hash


def register_patches():
    # TODO: PYT-797 pycryptodomex patches are exactly the same, except they have Cryptodome as
    # the top-level module name. It's possible that Crypto calls Cryptodome under the
    # hood, and so we end up patching both. More investigation required.
    register_post_import_hook(build_hash_patch("MD2Hash"), "Crypto.Hash.MD2")
    register_post_import_hook(build_hash_patch("MD4Hash"), "Crypto.Hash.MD4")
    register_post_import_hook(build_hash_patch("MD5Hash"), "Crypto.Hash.MD5")
    register_post_import_hook(build_hash_patch("SHA1Hash"), "Crypto.Hash.SHA1")


def reverse_patches():
    crypto_module = sys.modules.get("Crypto")
    if not crypto_module:
        return

    patch_manager.reverse_patches_by_owner(crypto_module.Hash.MD2.MD2Hash)
    patch_manager.reverse_patches_by_owner(crypto_module.Hash.MD4.MD4Hash)
    patch_manager.reverse_patches_by_owner(crypto_module.Hash.MD5.MD5Hash)
    patch_manager.reverse_patches_by_owner(crypto_module.Hash.SHA1.SHA1Hash)



# -*- coding: utf-8 -*-
# Copyright © 2021 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
"""
Definitions of view functions that can be used in django framework tests.
"""
import base64
import csv
import io
import json
import os
import pickle
from xml.etree import ElementTree


import lxml.etree
import six  # pylint: disable=did-not-import-extern
from Crypto.Cipher import Blowfish

import django
from django.contrib.auth import logout as DjangoLogout
from django.core.files.base import ContentFile
from django.core.files.storage import default_storage
from django.http import HttpResponse, StreamingHttpResponse
from django.shortcuts import redirect, render
from django.views.decorators.http import require_http_methods
from django.views.decorators.csrf import csrf_exempt

from contrast.test.mongo import create_mongo_client

from contrast.test.library_analysis import (
    remove_sys_module_entries,
    import_with_relative_imports,
    import_namespace_package,
    import_sample_package_onefile,
    import_multiple_tlds,
)

from vulnpy.trigger import cmdi

try:
    from django.urls import re_path as compat_url
except ImportError:
    from django.conf.urls import url as compat_url

try:
    from app.models import Message
except ImportError:
    Message = None

from app import settings

try:
    from app.jinja2_env import environment as jinja2_environment

    JINJA_ENV = jinja2_environment()
except ImportError:
    JINJA_ENV = None


# This ends up resolving to a path relative to the framework_test dir so that we don't
# have to package this data file with the contrast module.
DATA_DIR = os.path.join("..", "data")


def home(request):
    return render(request, "base.html")


def static_image(request):
    filename = request.GET.get("filename", "image.jpg")
    image_path = os.path.join(DATA_DIR, filename)
    with open(image_path, "rb") as image_fh:
        return StreamingHttpResponse(image_fh.readlines(), content_type="image/jpeg")


def django_autoescape(request):
    """Uses a django template with autoescaping (enabled by default)"""
    user_input = request.GET.get("user_input", "")
    return render(request, "xss.html", {"user_input": user_input}, using="django")


def django_autoescape_safe(request):
    """Uses a django template with autoescaping but marks value as safe"""
    user_input = request.GET.get("user_input", "")
    return render(
        request, "xss_autoescape_safe.html", {"user_input": user_input}, using="django"
    )


def django_no_autoescape(request):
    user_input = request.GET.get("user_input", "")
    return render(
        request, "xss_no_autoescape.html", {"user_input": user_input}, using="django"
    )


@require_http_methods(["POST"])
def file_contents(request):
    """
    This view is intended to demonstrate that simply calling request.FILES
    tracks the contents of a file because
    django.utils.datastructures.MultiValueDict __getitem__ is a source in policy.json
    """
    request.FILES["upload"]
    return render(request, "base.html")


def uce_base64(request):
    """
    Adapted from a vulnerable view in lets-be-bad-guys.
    The user_input param must base64 encoded string.
    Example: X19pbXBvcnRfXygnb3MnKS5zeXN0ZW0oJ2VjaG8gaGFja2VkJyk%3D
    this decodes to "__import__('os').system('echo hacked')"
    Note that in the example, UCE can lead to CMDi!
    """
    user_input = six.ensure_str(request.GET.get("user_input", None))

    # provide altchars to ensure a call to str.translate()
    decoded_input = base64.b64decode(user_input, altchars="-_")

    eval(decoded_input)

    return render(
        request,
        "unsafe_code_exec.html",
        {
            "result": "decoded input: {}".format(decoded_input),
            "user_input": "user input not displayed to prevent XSS",
        },
    )


def nosqli(request):
    if request.method == "GET":
        return render(request, "nosqli.html")
    if request.method == "POST":
        user_input = request.POST.get("user_input")
        with_kwarg = request.POST.get("with_kwarg", str(False)) == str(True)
        if user_input:
            try:
                method = request.POST["method_to_test"]
            except KeyError:
                raise Exception(
                    "Must indicate which pymongo collection method to test."
                )

            data = run_pymongo(method, user_input, with_kwarg)
            result = [x for x in data]
        else:
            result = "no user input"

    return render(request, "nosqli.html", {"result": result})


def run_pymongo(method, user_input, with_kwarg):
    """
    :param method: Pymongo Collection method to run
    :param user_input: input to pass in to method to run
    :return: any data returned from pymongo query
    """
    client = create_mongo_client()
    db = client.pymongo_test_db

    data = []
    if method == "find":
        data = (
            db.posts.find(filter={"title": user_input})
            if with_kwarg
            else db.posts.find({"title": user_input})
        )
    elif method == "insert_one":
        new_record = json.loads(user_input)
        db.posts.insert_one(document=new_record) if with_kwarg else db.posts.insert_one(
            new_record
        )
    elif method == "insert_many":
        if with_kwarg:
            db.posts.insert_many(documents=[{"title": user_input}])
        else:
            db.posts.insert_many([{"title": user_input}])
    elif method == "update":
        record = {"title": "Old title", "content": "PyMongo is fun!", "author": "Dani"}
        db.posts.insert_one(record)
        db.posts.update_one(record, {"$set": {"title": user_input}})
    elif method == "delete":
        db.posts.delete_one({"title": user_input})
    else:
        raise Exception("Pymongo query for method {} not implemented".format(method))

    return data


def xss(request):
    user_input = request.GET.get("user_input", "")
    should_escape = request.GET.get("django_escape", "False") == "True"

    if should_escape:
        from django.utils.html import escape

        user_input = escape(user_input)
    return render(request, "xss.html", {"user_input": user_input})


def raw_xss_streaming(request):
    user_input = request.GET.get("user_input", "")
    return StreamingHttpResponse(["<p>Looks like xss: " + user_input + "</p>"])


def xss_csv(request):
    user_input = request.GET.get("user_input", "")
    response = HttpResponse(content_type="text/csv")

    # write a csv-like response because agent currently does not support propagation thru the csv module
    response.write(",".join(["First row", user_input, "Bar", "Baz"]))
    return response


def render_to_response(filename, context=None):
    if context is None:
        context = {}
    template = JINJA_ENV.get_template(filename)
    return HttpResponse(template.render(**context))


def xss_jinja(request):
    user_input = request.GET.get("user_input", "")
    return render_to_response("xss.html", {"user_input": user_input})


def xss_mako(request):
    user_input = request.GET.get("user_input", "")

    return render(
        template_name="xss.html",
        context={"user_input": user_input},
        request=request,
        using="mako",
    )


def xss_cookie(request):
    cookie = request.COOKIES["user_input"]
    return render(request, "xss.html", {"user_input": cookie})


@require_http_methods(["POST"])
@csrf_exempt
def cmdi_file(request):
    """
    This view is intended to replicate (more-or-less) the SQLi vuln in djangoat
    """
    uploaded_file = request.FILES["upload"]
    data_path = os.path.join(settings.MEDIA_ROOT, "data")

    # Tests string propagation through path methods
    full_file_name = os.path.join(data_path, uploaded_file.name)
    content = ContentFile(uploaded_file.read())
    default_storage.save(full_file_name, content)

    bak_file_path = full_file_name + ".bak"
    # This is really cheesy isn't it?
    cmdi.do_os_system("cp {} {}".format(full_file_name, bak_file_path))

    return render(request, "cmdi.html")


@require_http_methods(["GET"])
def stream_source(request):
    """
    Emulates the case where a stream is marked as a source
    """
    user_input = b"\x80\x03]q\x00(K\x01K\x02K\x03e."  # [1, 2, 3]
    stream = io.BytesIO(user_input)

    try:
        # This is artificial and would never happen inside a real app, but it
        # allows us to test the case where a stream is treated as a source
        stream.cs__source = True
        result = pickle.load(stream)
    except Exception as e:
        result = str(e)

    return render(
        request, "deserialization.html", {"user_input": "ud", "result": result}
    )


def unvalidated_redirect(request):
    user_input = request.GET.get("user_input", "/cmdi/")
    permanent = request.GET.get("permanent", False)
    return redirect(user_input, permanent=permanent)


@require_http_methods(["GET"])
def two_vulns(request):
    """
    This view contains two vulnerabilities.
    """
    user_input = request.GET.get("user_input", "")

    try:
        cmdi.do_subprocess_popen(user_input)
    except Exception:
        pass

    # Second vulnerability: reflected-xss
    return HttpResponse("<p>{}</p>".format(user_input))


def xpath_injection(request):
    query = request.GET.get("query", "")
    module = request.GET.get("module", "")

    xml_doc = """<foo><bar name='whatever'>42</bar></foo>"""

    result = None

    try:
        if module == "xml":
            node = ElementTree.fromstring(xml_doc)
            result = node.find(query)
        elif module == "lxml":
            node = lxml.etree.fromstring(xml_doc)
            result = node.find(query)
    except Exception as e:
        result = str(e)

    return HttpResponse("<p>result = {}</p>".format(result))


def logout(request):
    DjangoLogout(request)
    return render(request, "logout.html")


def get_dynamic_source(request, request_dict, source):
    user_input = ""

    if source == "parameter":
        user_input = request_dict.get("user_input")
    elif source == "body":
        user_input = request.body
    elif source == "host":
        user_input = request.get_host()
    elif source == "port":
        user_input = request.get_port()
    elif source == "raw_uri":
        user_input = request.get_raw_uri()
    elif source == "files" and request.method == "POST":
        file_stream = request.FILES.get("user_input")
        user_input = file_stream.read()
    elif source == "full_path":
        user_input = request.get_full_path()
    elif source == "full_path_info":
        user_input = request.get_full_path_info()
    elif source == "scheme":
        user_input = request.scheme
    elif source == "encoding":
        user_input = request.encoding
    elif source == "referer_header":
        user_input = (
            request.headers["Referer"]
            if django.VERSION >= (2, 2)
            else request.META["HTTP_REFERER"]
        )
    elif source == "http_method":
        user_input = request.method

    return user_input


def dynamic_sources(request):
    request_dict = request.POST if request.method == "POST" else request.GET
    source = request_dict.get("source", "")
    user_input = get_dynamic_source(request, request_dict, source)

    if isinstance(user_input, bytes):
        # TODO: PYT-714 be able to remove this conversion when .format can propagate tags
        user_input = str(user_input)

    cmdi.do_os_system("echo {}".format(user_input))

    return HttpResponse("<p>tested sources </p>")


def cookie_source(request):
    user_input = request.COOKIES["user_input"]
    cmdi.do_os_system("echo {}".format(user_input))

    return HttpResponse("<p>echoed cookie: {}</p>".format(user_input))


def header_source(request):
    use_environ = request.environ.get("use_environ", False)

    if use_environ:
        user_input = request.environ.get("HTTP_TEST_HEADER")
    else:
        user_input = (
            request.headers["Test-Header"]
            if django.VERSION >= (2, 2)
            else request.META["HTTP_TEST_HEADER"]
        )

    cmdi.do_os_system("echo {}".format(user_input))

    return HttpResponse("<p>echoed header: {}</p>".format(user_input))


def header_key_source(request):
    use_environ = request.environ.get("use_environ", False)

    if use_environ:
        source_dict = request.environ
    else:
        source_dict = request.headers if django.VERSION >= (2, 2) else request.META

    # roundabout way of getting the key we want out of the dictionary without changing its id
    user_input = [
        k for k in source_dict.keys() if k in ("HTTP_TEST_HEADER", "Test-Header")
    ][0]

    cmdi.do_os_system("echo {}".format(user_input))

    return HttpResponse("<p>echoed header key: {}</p>".format(user_input))


def stored_xss(request):
    user_input = request.GET.get("user_input", "hello")
    test_type = request.GET.get("test_type")

    if test_type == "create_method":
        msg = Message.objects.create(name=user_input)
    elif test_type == "init_direct":
        msg = Message(name=user_input)
        msg.save()
    elif test_type == "filter":
        _ = Message.objects.create(name=user_input)
        msgs = Message.objects.filter(name=user_input)
        msg = msgs[0]
    elif test_type == "multi-column":
        msg = Message.objects.create(name=user_input, content=user_input)
    else:
        raise Exception("Pass a test_type to test")

    html = "<p>Looks like xss: " + msg.name + "</p>"
    return HttpResponse(html)


def trust_boundary_violation(request):
    user_input = request.GET.get("user_input")
    setdefault = request.GET.get("setdefault") == "True"

    if setdefault:
        request.session.setdefault("whatever", user_input)
    else:
        request.session["whatever"] = user_input

    return HttpResponse("<p>Trust boundary violation</p>")


def crypto_bad_ciphers(request):
    cipher_result = Blowfish.new(b"example_len16key", mode=Blowfish.MODE_EAX)

    return render(request, "crypto_bad_ciphers.html", {"result": str(cipher_result)})


class Echo:
    def write(self, value):
        return value


def streaming(request):
    # Taken from https://docs.djangoproject.com/en/2.0/howto/outputting-csv/#streaming-large-csv-files
    if request.method == "GET":
        attack = request.GET.get("attack", False)
        if attack:
            rows = (["Row {}".format(idx), str(idx)] for idx in range(65536))
            pseudo_buffer = Echo()
            writer = csv.writer(pseudo_buffer)
            response = StreamingHttpResponse(
                (writer.writerow(row) for row in rows), content_type="text/csv"
            )
            response[
                "Content-Disposition"
            ] = 'attachment; filename="runescapetrivia.csv"'
            return response
        return render(request, "streaming.html", {"result": ""})
    return render(request, "streaming.html", {"result": "Try a GET Request instead."})


def new_dynamic(request):
    return render(request, "base.html")


def other_dynamic(request):
    return render(request, "base.html")


def dynamic_url_view(request):
    from app.urls import urlpatterns

    user_input = request.GET.get("user_input", "new_dynamic")

    func = new_dynamic if user_input == "new_dynamic" else other_dynamic
    urlpatterns.append(compat_url(r"^{}/?$".format(user_input), func))
    return render(request, "base.html")


def nosqli_seed(request):
    client = create_mongo_client()
    db = client.pymongo_test_db
    posts = db.posts
    post_1 = {
        "id": "Python and MongoDB",
        "content": "PyMongo is fun, you guys",
        "author": "Scott",
    }
    post_2 = {
        "title": "Virtual Environments",
        "content": "Use virtual environments, you guys",
        "author": "Scott",
    }
    post_3 = {
        "title": "Learning Python",
        "content": "Learn Python, it is easy",
        "author": "Bill",
    }
    posts.insert_many([post_1, post_2, post_3])
    return render(request, "nosqli.html", {"error": "DB Seeded."})


def import_package_with_relative_imports(request):
    rm_modules = request.GET.get("rm_sys_mod_entries", None)

    if rm_modules:
        remove_sys_module_entries()

    import_with_relative_imports()

    return HttpResponse("")


def import_sample_namespace_package(request):
    # Deleting top level namespace modules breaks them in py2.
    # Namespace name is cached in sys modules, but shows up
    # as a builtin module in py2.
    import_namespace_package()

    return HttpResponse("")


def import_package_onefile(request):
    rm_modules = request.GET.get("rm_sys_mod_entries", None)

    if rm_modules:
        remove_sys_module_entries()

    import_sample_package_onefile()

    return HttpResponse("")


def import_sample_dist_multiple_tlds(request):
    rm_modules = request.GET.get("rm_sys_mod_entries", None)

    if rm_modules:
        remove_sys_module_entries()

    import_multiple_tlds()

    return HttpResponse("")



# -*- coding: utf-8 -*-
# Copyright © 2021 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
"""
Common code for test applications
"""


def getitem_request(request, key, default=""):
    if request.method == "GET":
        param = request.GET
    elif request.method == "POST":
        param = request.POST
    else:
        return default

    try:
        return param[key]
    except KeyError:
        return default



# -*- coding: utf-8 -*-
# Copyright © 2021 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.extern.six import iteritems, PY2, PY3

import contrast
from contrast.agent.assess.contrast_event import ContrastEvent
from contrast.agent.policy.loader import Policy
from contrast.agent.assess.policy.propagation_node import PropagationNode
from contrast.agent.assess.policy.propagators import (
    IGNORE_LENGTH_ACTIONS,
    PROPAGATOR_ACTIONS,
    STREAM_ACTIONS,
    TAG_ACTION,
    BasePropagator,
    stream_propagator,
)
from contrast.agent.assess.policy.source_policy import SourceNode, apply_stream_source
from contrast.agent.assess.properties import Properties
from contrast.agent.assess.policy.preshift import Preshift
from contrast.agent.assess.utils import (
    copy_tags_to_offset,
    copy_events,
    copy_from,
    is_tracked,
    get_properties,
    set_properties,
    track_string,
)
from contrast.utils.assess.duck_utils import (
    is_filelike,
    is_iterable,
    safe_getattr,
    safe_iterator,
    len_or_zero,
)
from contrast.agent.settings_state import SettingsState

from contrast.extern import structlog as logging

logger = logging.getLogger("contrast")


def apply(nodes, preshift, ret):  # pylint: disable=redefined-builtin
    if not nodes:
        return

    for node in nodes:
        if not preshift:
            continue

        target = node.get_matching_first_target(
            preshift.obj, ret, preshift.args, preshift.kwargs
        )

        apply_propagator(node, preshift, target, ret, None)


def is_target_valid_length(target, action):
    if action in IGNORE_LENGTH_ACTIONS:
        return True

    return len_or_zero(target) > 0


def track_copy_without_new_event(target, self_obj):
    """
    In general, when a string propagation event results in a copy of the
    original object we track the newly created copy but don't record an event.
    This behavior might be modified in the future.

    Usually, python is smart enough not to create a new string object that is
    simply a copy of the original. However, there are several exceptions to
    this rule.

    This method should only be invoked in these special cases, and they should
    be documented heavily.
    """
    if is_tracked(self_obj):
        track_string(target)
        copy_from(target, self_obj, 0, set())


STREAM_SOURCES = ["read", "read1", "readline", "readlines", "getvalue"]
STREAM_WRITE_METHODS = ["write", "writelines"]


def create_stream_event(node_type, stream, frame, args, kwargs):
    context = contrast.CS__CONTEXT_TRACKER.current()
    if context is None or not SettingsState().is_assess_enabled():
        return None

    module = stream.__class__.__module__
    class_name = stream.__class__.__name__

    if not args and not kwargs:
        # We add an empty string if an empty stream is created for reporting purposes
        args = ("",)

    source_type = safe_getattr(stream, "cs__source_type", None) or "BODY"

    if node_type == "source":
        node = SourceNode(
            module,
            class_name,
            True,
            "__init__",
            "ARG_0,KWARG:initial_value",
            source_type,
            None,
        )
    else:
        node = PropagationNode(
            module,
            class_name,
            True,
            "__init__",
            "ARG_0,KWARG:initial_value",
            "RETURN",
            source_type,
            None,
        )

    return ContrastEvent(
        node, stream, stream, None, args, kwargs, [], 0, None, frame=frame
    )


def create_stream_source_event(stream, frame, args, kwargs):
    """
    Called directly from C extensions to create source events for __init__
    """
    init_event = create_stream_event("source", stream, frame, args, kwargs)
    stream.cs__source_event = init_event

    if len(args) > 0:
        properties = get_properties(args[0])
        if properties is not None:
            stream_props = Properties(stream)
            set_properties(stream, stream_props)
            copy_tags_to_offset(stream_props, properties.tags, 0)
            copy_events(stream_props, properties)
            prop_event = create_stream_event("propagation", stream, frame, args, kwargs)
            if prop_event is not None:
                if len(properties.events) > 0:
                    prop_event.parent_ids = [properties.events[-1].event_id]
                stream_props.add_event(prop_event)
            stream.cs__tracked = True

    # TODO: PYT-1405 support kwargs


def propagate_stream(method_name, target, self_obj, ret, frame, args, kwargs):
    """
    Called directly from C extensions to propagate stream operations.
    """
    args = [] if args is None else args
    kwargs = {} if kwargs is None else kwargs
    preshift = Preshift(self_obj, args, kwargs)

    if method_name in STREAM_WRITE_METHODS:
        stream_propagator.propagate_stream_write(
            method_name, preshift, target, args, kwargs
        )

    elif (
        self_obj.cs__tracked
        and get_properties(self_obj)
        and get_properties(self_obj).tags
    ):
        propagation_method = STREAM_ACTIONS.get(method_name)
        if propagation_method is None:
            return

        propagation_method(method_name, preshift, target, ret, frame)

    # If the stream is already considered tracked, it will not be treated as a
    # source.
    elif self_obj.cs__source and method_name in STREAM_SOURCES:
        apply_stream_source(method_name, target, self_obj, ret, args, kwargs)


def propagate(method_name, target, self_obj, ret, frame, args, kwargs):
    """Called directly from C extensions to propagate string operations"""
    if not isinstance(self_obj, bytearray) and ret is self_obj:
        return

    if is_filelike(self_obj) and self_obj.closed:
        return

    if cast_special_case(method_name, args, target):
        return

    # Copies of strings (unicode) and bytes objects usually return the same
    # object. However, copies of bytearrays return a new object, so it is
    # necessary for us to track the new string, even if it hasn't changed.
    if isinstance(self_obj, bytearray) and ret == self_obj and method_name != "concat":
        track_copy_without_new_event(target, self_obj)
        return

    if translate_special_case(method_name, self_obj, ret):
        track_copy_without_new_event(target, self_obj)
        return

    # Account for the fact that None is a valid argument for cformat, and it
    # also may be a single object. The preshift requires an argument arrary.
    # Eventually this should maybe just move into the C hook for cformat,
    # although it is a bit easier to implement (and debug) where it is here
    if method_name == "cformat" and not isinstance(args, tuple):
        args = (args,)

    args = [] if args is None else args
    kwargs = {} if kwargs is None else kwargs
    preshift = Preshift(self_obj, args, kwargs)

    policy = Policy()

    for node in policy.propagators_by_name["BUILTIN.str." + method_name]:
        apply_propagator(node, preshift, target, ret, frame)
        break


# These methods return lists and are handled by the SPLIT propagation action.
# We don't want to iterate over the lists during propagation, but instead need
# to make sure we call the right propagation action first.
SPLIT_PROPAGATOR_METHODS = ["split", "rsplit", "splitlines", "partition", "rpartition"]
REGEX_MODULE = "re"
REGEX_SPLIT = "split"
REGEX_FINDALL = "findall"


def cast_special_case(method_name, args, target):
    """
    We don't want to record an event when `target is source`.

    This will never happen for bytearrays, because they are always at least copied.
    For bytearrays, we check if the source and target are ==, and if so we copy
    tags over without recording a new event.

    It would be better if there were some way to check for this slightly later,
    based on the policy node's information.

    Currently, every other string function is an instance method, so we short-
    circuit out of propagation if `self_obj is ret`. For casting, self_obj is None
    because casting (which is really __new__ for unicode/bytes) is a module method
    """
    if method_name == "CAST":
        source = args and args[0]
        if target is source:
            return True
        if (
            isinstance(source, bytearray)
            and isinstance(target, bytearray)
            and target == source
        ):
            track_copy_without_new_event(target, source)
            return True
    return False


def translate_special_case(method_name, self_obj, ret):
    """
    py27 unicode objects and py3x str objects returned by str.translate()
    are always new objects, so in order to keep our behavior consistent we
    do not create a new event in the case where new == original

    If this method returns True, we should copy all tags from
    self_obj to target without creating a new event.
    """
    return (
        method_name == "translate"
        and ret == self_obj
        and (
            (PY2 and isinstance(self_obj, unicode))
            or (PY3 and isinstance(self_obj, str))
        )
    )


def has_iterable_retval(propagator_node):
    """
    Returns True if given propagator node expects iterable list/tuple return value

    Ordinarily, if we encounter a return value that is a list/tuple, we recurse down in
    order to find simple strings. However, some propagators expect the return value to
    be in an iterable, in which case we need to use the iterable itself to build the
    preshift.
    """
    if propagator_node.module == REGEX_MODULE and propagator_node.method_name in [
        REGEX_SPLIT,
        REGEX_FINDALL,
    ]:
        return True
    if propagator_node.class_name != "str":
        return False
    return propagator_node.method_name in SPLIT_PROPAGATOR_METHODS


def apply_propagator(propagator_node, preshift, target, ret, frame):
    if not propagator_node or not target:
        return

    if not is_target_valid_length(target, propagator_node.action):
        return

    if isinstance(target, dict):
        for key, value in iteritems(target):
            apply_propagator(propagator_node, preshift, key, ret, frame)
            apply_propagator(propagator_node, preshift, value, ret, frame)
    elif is_iterable(target) and not has_iterable_retval(propagator_node):
        for item in safe_iterator(target):
            apply_propagator(propagator_node, preshift, item, ret, frame)
    else:
        propagate_string(propagator_node, preshift, target, ret, frame)


def propagate_string(propagator_node, preshift, target, ret, frame):
    action = propagator_node.action
    propagator_class = PROPAGATOR_ACTIONS.get(action, BasePropagator)
    if propagator_class is BasePropagator and action != TAG_ACTION:
        logger.warning(
            "Unknown action type %s. Using default propagation for %s.%s.",
            propagator_node.action,
            propagator_node.location,
            propagator_node.method_name,
        )

    propagator = propagator_class(propagator_node, preshift, target)

    if not propagator.needs_propagation:
        return

    propagator.track_and_propagate(ret, frame)
    logger.debug(
        "Propagator %s found: propagated to %s", propagator_node.name, id(target)
    )



# -*- coding: utf-8 -*-
# Copyright © 2021 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import time

from contrast.extern.six import iteritems

from contrast.agent.settings_state import SettingsState
from contrast.api.settings_pb2 import (
    AgentSettings,
    ProtectionRule,
)

from contrast.configuration import AgentConfig

from contrast.extern import structlog as logging

logger = logging.getLogger("contrast")


class SettingsBuilder(object):
    DEFAULT_FEATURES = {
        "defend": {"enabled": True},
        "log_file": "contrast_test.log",
        "log_level": "INFO",
    }

    DEFAULT_SETTINGS = {
        "protection_rules": [
            {
                "id": "cmd-injection",
                "name": "Command Injection",
                "mode": ProtectionRule.BLOCK,
            },
            {
                "id": "method-tampering",
                "name": "Http Method Tampering",
                "mode": ProtectionRule.BLOCK,
            },
            {
                "id": "path-traversal",
                "name": "Path Traversal",
                "mode": ProtectionRule.BLOCK,
            },
            {
                "id": "sql-injection",
                "name": "SQL Injection",
                "mode": ProtectionRule.BLOCK,
            },
            {
                "id": "reflected-xss",
                "name": "Cross-Site Scripting",
                "mode": ProtectionRule.BLOCK,
            },
            {
                "id": "untrusted-deserialization",
                "name": "Untrusted Deserialization",
                "mode": ProtectionRule.BLOCK,
            },
            {
                "id": "xxe",
                "name": "XML External Entity Processing",
                "mode": ProtectionRule.BLOCK,
            },
            {
                "id": "reflected-xss",
                "name": "Cross-Site Scripting",
                "mode": ProtectionRule.BLOCK_AT_PERIMETER,
            },
            {
                "id": "unsafe-code-execution",
                "name": "Unsafe Code Execution",
                "mode": ProtectionRule.BLOCK,
            },
            {
                "id": "ssrf",
                "name": "Server-Side Request Forgery",
                "mode": ProtectionRule.BLOCK,
            },
            {
                "id": "unsafe-file-upload",
                "name": "Server-Side Request Forgery",
                "mode": ProtectionRule.BLOCK_AT_PERIMETER,
            },
            {
                "id": "nosql-injection",
                "name": "Command Injection",
                "mode": ProtectionRule.BLOCK,
            },
        ],
        "disabled_assess_rules": ["test_application_disabled"],
    }

    def __init__(self):
        self._settings = None

    def settings(self, settings_dict=None, features_dict=None, defend_rules=None):
        self._settings = SettingsState()
        self._settings.config = AgentConfig()

        if settings_dict or features_dict or defend_rules:
            agent_settings = self.service_response(
                settings_dict=self._defend_rule_settings(defend_rules),
                features_dict=features_dict,
            )
        else:

            agent_settings = self.service_response(
                settings_dict=self.DEFAULT_SETTINGS, features_dict=self.DEFAULT_FEATURES
            )

        self._settings.process_service_response(agent_settings)

        self._settings.set_protect_rules()

        if defend_rules:
            needed_rules = [dr["id"] for dr in defend_rules]

            self._settings.defend_rules = {
                k: v
                for k, v in iteritems(self._settings.defend_rules)
                if k in needed_rules
            }

        return self._settings

    def build(self, settings_dict=None, features_dict=None, defend_rules=None):
        return self.settings(
            settings_dict=settings_dict,
            features_dict=features_dict,
            defend_rules=defend_rules,
        )

    def build_assess(self):
        settings = self.settings()
        settings.server_features.defend.enabled = False
        settings.server_features.assess.enabled = True

        settings.config.put("assess.enable", True)
        settings.config.put("protect.enable", False)

        return settings

    def service_response(self, settings_dict=None, features_dict=None):
        obj = AgentSettings()
        obj.sent_ms = int(time.time() * 1000)

        SettingsBuilder.populate_from_dict(
            obj.server_features,
            features_dict if features_dict else self.DEFAULT_FEATURES,
        )

        SettingsBuilder.populate_from_dict(
            obj.application_settings,
            settings_dict if settings_dict else self.DEFAULT_SETTINGS,
        )

        return obj

    def _defend_rule_settings(self, defend_rules=None):
        if defend_rules:
            return {
                "protection_rules": [
                    SettingsBuilder._defend_rule_entry(rule) for rule in defend_rules
                ]
            }

        return self.DEFAULT_SETTINGS

    @staticmethod
    def _defend_rule_entry(rule):
        if isinstance(rule, dict):
            return rule

        return {"id": str(rule), "name": str(rule), "mode": ProtectionRule.BLOCK}

    @staticmethod
    def populate_from_dict(pb, values):
        """
        Initialize the given protobuf instance from a python dictionary
        """
        for key, value in iteritems(values):
            if isinstance(value, dict):
                SettingsBuilder.populate_from_dict(getattr(pb, key), value)
            elif isinstance(value, list):
                SettingsBuilder.populate_from_list(getattr(pb, key), value)
            else:
                try:
                    setattr(pb, key, value)
                except AttributeError:
                    logger.error(
                        "Unable to set protobuf instance %s : %s=%s",
                        type(pb),
                        key,
                        value,
                    )

        return pb

    @staticmethod
    def populate_from_list(pb, values):
        """
        Initialize the given protobuf instance from a python list
        """
        if len(values) > 0:
            if isinstance(values[0], dict):
                for value in values:
                    child = pb.add()
                    SettingsBuilder.populate_from_dict(child, value)
            else:
                pb.extend(values)



# -*- coding: utf-8 -*-
# Copyright © 2021 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.api.dtm_pb2 import TraceEvent
from contrast.agent.policy import constants
from contrast.utils.base64_utils import base64_decode


def default_taint_range_validator(taint_ranges, endrange):
    for taint_range in taint_ranges:
        assert taint_range.range == "0:{}".format(endrange)


def _assert_basic_event(event, event_type, action, class_name, method_name):
    assert event.type == event_type
    assert event.action == action
    assert event.signature.class_name == class_name
    assert event.signature.method_name == method_name


def check_event(
    event,
    event_type,
    action,
    class_name,
    method_name,
    source_types,
    first_parent,
    source=None,
    target=None,
    ret_value=None,
    taint_range_validator=default_taint_range_validator,
):
    """

    Assert values for TraceEvent dtm.
    """
    _assert_basic_event(event, event_type, action, class_name, method_name)

    assert all([s.type in source_types for s in event.event_sources])

    assert (
        event.parent_object_ids[0].id == first_parent.object_id
        if first_parent
        else True
    )

    if source:
        assert event.source == source

    if target:
        assert event.target == target

    if ret_value:
        # ret value is the right side of the vulnerability > details page in TS
        assert base64_decode(event.ret.value) == ret_value
        if taint_range_validator is not None:
            taint_range_validator(list(event.taint_ranges), len(ret_value))


def assert_source_event(event, class_name, method_name):
    _assert_basic_event(
        event,
        TraceEvent.TYPE_PROPAGATION,
        TraceEvent.Action.Value(constants.CREATION_TYPE),
        class_name,
        method_name,
    )


def assert_trigger_event(event, class_name, method_name):
    _assert_basic_event(
        event,
        TraceEvent.TYPE_PROPAGATION,
        TraceEvent.Action.Value(constants.TRIGGER_TYPE),
        class_name,
        method_name,
    )


def assert_propagation_event(event, class_name, method_name, a2r=True):
    prop_action = "{}{}{}".format(
        constants.ALL_TYPE if a2r else constants.OBJECT_KEY,
        constants.TO_MARKER,
        constants.RETURN_KEY,
    )

    _assert_basic_event(
        event,
        TraceEvent.TYPE_PROPAGATION,
        TraceEvent.Action.Value(prop_action),
        class_name,
        method_name,
    )



# -*- coding: utf-8 -*-
# Copyright © 2021 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.extern import six

from contrast.extern import structlog as logging

logger = logging.getLogger("contrast")


def truncate(value, default=""):
    """
    Truncate to 256 characters
    """
    if value is None:
        return default

    return value[:256]


def truncated_signature(value):
    """
    Get a log-friendly representation of a potentially long string. This function
    will truncate the string if necessary.

    First, we truncate the input to 60 characters - if this does happen, we'll also add
    `[TRUNCATED]` to the output. We then append the string's `id`. The string is
    converted to its most readable form using __repr__, which means that any newlines
    or invisible chars (like BEL) will be converted to something nice and readable.

    examples:
    'This is some string' (id=4443462824)
    'Here is a very long string that is longer th' [TRUNCATED] (id=4443294816)

    :param value: string whose truncated signature we want
    :return: string representation of the input value, truncated to 60 chars with
        its `id` appended as well. On any failure, return only the id.
    """
    try:
        append_truncate = ""

        if isinstance(value, bytearray):
            value = bytes(value)

        if len(value) > 60:
            value = value[:60]
            append_truncate = " [TRUNCATED]"

        value = six.ensure_str(value, errors="ignore")
        return "{!r}{} (id={})".format(value, append_truncate, id(value))
    except Exception as e:
        logger.debug("Failed to truncate string: %s", e)
        return "(id={})".format(id(value))


def index_of_any(value, search_chars):
    """
    Find the first index of a char in a string
    :param value: string to search
    :param search_chars: strings to search for
    :return: index if found else -1
    """

    for sc in search_chars:

        index = value.find(sc)

        if index != -1:
            return index

    return -1


def ends_with_any(value, strings):
    """
    Returns True if any of the strings are at the end of the value
    """
    return any(value.endswith(item) for item in strings)


def equals_ignore_case(this, that):
    return this.lower() == that.lower()


def protobuf_safe(value):
    return "" if value is None else value


def ensure_string(value):
    return value if isinstance(value, six.string_types) else value.decode("utf-8")



# -*- coding: utf-8 -*-
# Copyright © 2021 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import sys

import contrast

from contrast.extern.wrapt import register_post_import_hook
from contrast.agent import scope
from contrast.agent.policy import patch_manager
from contrast.agent.assess.utils import copy_from, track_string

from contrast.utils.patch_utils import patch_cls_or_instance

from contrast.extern import structlog as logging

logger = logging.getLogger("contrast")

SAFESTRING_MODULE_NAME = "django.utils.safestring"
MARK_SAFE_NAME = "mark_safe"


def _get_source(args, kwargs):
    if args:
        return args[0]
    if kwargs:
        return kwargs.get("s")
    return None


def mark_safe_patch(orig_func, patch_policy, *args, **kwargs):
    """
    This patch implements a more optimized "deadzoned" original mark_safe call.

    We are deadzoning calling the original mark_safe because it may call SafeText,
    a class that inherits from str and that ends up propagating excessively.

    Because mark_safe is called many times when Django renders a template,
    excessive propagation is incredibly costly. So instead, we do the same work as
    the KEEP propagator would, but with a deadzoned call to the original function.
    """
    # if we're already in scope, don't bother doing any analysis
    if scope.in_scope():
        return orig_func(*args, **kwargs)

    # if we're not yet in scope, call orig_func and analysis in scope
    with scope.contrast_scope():
        # We don't wrap orig_func call in try/catch because if it raises an error
        # we don't want to do any analysis.
        result = orig_func(*args, **kwargs)

        if contrast.CS__CONTEXT_TRACKER.current() is not None:
            try:
                logger.debug("Analyzing in %s custom propagator.", MARK_SAFE_NAME)
                source = _get_source(args, kwargs)
                track_string(result)
                copy_from(result, source)
            except Exception as e:
                logger.debug(
                    "Failed to analyse in %s propagator. %s", MARK_SAFE_NAME, str(e)
                )

    return result


def patch_django(safestring_module):
    patch_cls_or_instance(safestring_module, MARK_SAFE_NAME, mark_safe_patch)


def register_patches():
    register_post_import_hook(patch_django, SAFESTRING_MODULE_NAME)


def reverse_patches():
    safestring_module = sys.modules.get(SAFESTRING_MODULE_NAME)
    if not safestring_module:
        return

    patch_manager.reverse_patches_by_owner(safestring_module)



# -*- coding: utf-8 -*-
# Copyright © 2021 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.api.dtm_pb2 import ArchitectureComponent


def append_db(activity, db_info):
    """
    Add an Architecture Component to Activity
    for Teamserver to display database usage.
    """
    if not db_info or not isinstance(db_info, dict):
        return

    arch_comp = _build_architecture_component(db_info)
    activity.architectures.extend([arch_comp])


def _build_architecture_component(obj):
    """
    Builds ArchitectureComponent from dict
    """
    arch_comp = ArchitectureComponent()

    # vendor must be a string that exactly matches a value from
    # Teamserver's flowmap/technologies.json > service > one of "name"
    if "vendor" in obj and obj["vendor"]:
        arch_comp.vendor = obj["vendor"]

    if "host" in obj and obj["host"]:
        arch_comp.remote_host = obj["host"]

    if "port" in obj and obj["port"]:
        port = obj["port"]
        arch_comp.remote_port = int(port) if port else -1

    arch_comp.type = "db"

    arch_comp.url = obj["database"] if "database" in obj else "default"

    return arch_comp



# -*- coding: utf-8 -*-
# Copyright © 2021 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
class ContrastServiceException(Exception):
    def __init__(self):
        message = "Unable to connect or send messages to Contrast Service."
        super(ContrastServiceException, self).__init__(message)



# -*- coding: utf-8 -*-
# Copyright © 2021 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import inspect
import sys
from contrast.extern.wrapt import register_post_import_hook
from contrast.agent.policy import patch_manager
from contrast.agent.assess.policy import propagation_policy
from contrast.utils.assess.stream_utils import BaseStreamProxy, stream_patch_body
from contrast.utils.patch_utils import patch_cls_or_instance

from contrast.extern import structlog as logging

logger = logging.getLogger("contrast")

CSTRINGIO_MODULE = "cStringIO"


def build_patch(self, orig_func, method_name):
    def patched_func(*args, **kwargs):
        return stream_patch_body(method_name, orig_func, self, *args, **kwargs)

    patched_func.__name__ = method_name
    return patched_func


class StringIOProxy(BaseStreamProxy):
    """
    Base class for stream types from the cStringIO module (Py2 only)
    """

    CS__METHOD_NAMES = BaseStreamProxy.CS__METHOD_NAMES + ["getvalue"]

    getvalue = None

    def __init__(self, wrapped):
        super(StringIOProxy, self).__init__(wrapped)
        for method_name in self.CS__METHOD_NAMES:
            orig_func = getattr(self, method_name)
            new_func = build_patch(self, orig_func, method_name)
            setattr(self, method_name, new_func)


class InputType(StringIOProxy):
    """Proxy class for cStringIO.InputType"""


class OutputType(StringIOProxy):
    """Proxy class for CstringIO.OutputType"""

    CS__METHOD_NAMES = StringIOProxy.CS__METHOD_NAMES + ["write", "writelines"]

    write = None
    writelines = None


def wrap_stream(stream):
    from cStringIO import InputType as OrigInputType

    return (
        InputType(stream) if isinstance(stream, OrigInputType) else OutputType(stream)
    )


def patched_StringIO(orig_func, patch_policy=None, *args, **kwargs):
    orig_stream = orig_func(*args, **kwargs)

    try:
        proxied_stream = wrap_stream(orig_stream)
        frame = inspect.currentframe()
        propagation_policy.create_stream_source_event(
            proxied_stream, frame.f_back, args, kwargs
        )
        return proxied_stream
    except Exception:
        logger.exception("Failed to propagate cStringIO.StringIO creation")

    return orig_stream


def patch_cStringIO(cStringIO_module):
    # cStringIO.StringIO is actually a factory function, not a class
    patch_cls_or_instance(cStringIO_module, "StringIO", patched_StringIO)


def register_patches():
    register_post_import_hook(patch_cStringIO, CSTRINGIO_MODULE)


def reverse_patches():
    stringio_module = sys.modules.get(CSTRINGIO_MODULE)
    if not stringio_module:
        return

    patch_manager.reverse_patches_by_owner(stringio_module)



# -*- coding: utf-8 -*-
# Copyright © 2021 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import os
import pytest

from pymongo import MongoClient


def mongo_db_exists():
    client = create_mongo_client(serverSelectionTimeoutMS=20)
    db = client.pymongo_test

    try:
        db.posts.count()
    except Exception as e:
        print("EXCEPTION", e)
        return False

    return True


def create_mongo_client(**kwargs):
    mongo_host = [os.environ.get("MONGODB_HOST", "localhost:27017")]
    return MongoClient(host=mongo_host, **kwargs)


skip_no_mongo_db = pytest.mark.skipif(
    not mongo_db_exists(), reason="Host does not have mongodb"
)



# -*- coding: utf-8 -*-
# Copyright © 2021 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import contrast
from contrast.agent import scope
from contrast.agent.policy.trigger_node import TriggerNode
from contrast.agent.middlewares.base_django_middleware import DjangoBaseMiddleware
from contrast.agent.middlewares.response_wrappers.django_response_wrapper import (
    DjangoResponseWrapper,
)
from contrast.agent.request_context import RequestContext
from contrast.utils.decorators import cached_property
from contrast.utils.exceptions.contrast_service_exception import (
    ContrastServiceException,
)

from contrast.extern import structlog as logging

logger = logging.getLogger("contrast")


class DjangoMiddleware(DjangoBaseMiddleware):
    """
    Newer style of Django Middleware; This middleware is for Django 1.10+ and 2.+.
    """

    setting_name = "MIDDLEWARE"

    def __init__(self, get_response=None):
        self.validate_middleware()
        super(DjangoMiddleware, self).__init__()

        self.get_response = get_response

    def __call__(self, request):
        if self.is_agent_enabled():
            body = self.get_body(request)
            if body is not None:
                context = RequestContext(request.environ, body)

                with contrast.CS__CONTEXT_TRACKER.lifespan(context):
                    return self.call_with_agent(context, request)

        return self.call_without_agent(request)

    def call_with_agent(self, context, request, environ=None):
        super(DjangoMiddleware, self).call_with_agent(context, request)

        try:
            self.prefilter(context)

            logger.debug("Start app code and get response")
            response = self.get_response(request)
            logger.debug("Finished app code and get response")

            with scope.contrast_scope():
                wrapped_response = DjangoResponseWrapper(response)

            context.extract_response_to_context(wrapped_response)

            self.postfilter(context)

            self.check_for_blocked(context)

            context.timer.set_end("total")

        except ContrastServiceException as e:
            logger.warning(e)
            return self.call_without_agent(request)
        except Exception as e:
            return self.handle_exception(e)
        finally:
            self.handle_ensure(context, request)
            self.log_end_request_analysis(request.path)
            if self.settings.is_assess_enabled():
                contrast.STRING_TRACKER.ageoff()

        return response

    def call_without_agent(self, request):
        """
        Normal without middleware call
        """
        super(DjangoMiddleware, self).call_without_agent()
        return self.get_response(request)

    @cached_property
    def trigger_node(self):
        """
        Trigger node property used by assess reflected xss postfilter rule
        """
        method_name = self.get_response.__name__

        module, class_name, args, instance_method = self._process_trigger_handler(
            self.get_response
        )

        return (
            TriggerNode(module, class_name, instance_method, method_name, "RETURN"),
            args,
        )



# -*- coding: utf-8 -*-
# Copyright © 2021 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import sys
from contrast.extern import six
from contrast.extern.wrapt import register_post_import_hook
from contrast.agent.policy import patch_manager
from contrast.assess_extensions import cs_str
from contrast.utils.patch_utils import patch_cls_or_instance

from contrast.extern import structlog as logging

logger = logging.getLogger("contrast")


START_METHOD = "start"
BOOTSTRAP_METHOD = "_bootstrap_inner" if six.PY3 else "_Thread__bootstrap_inner"


def start(orig_func, patch_policy, self, *args, **kwargs):
    import threading

    try:
        # Save the scope of the current (creating) thread to copy to new thread
        self.cs__parent_scope = cs_str.get_thread_scope()
        # This is used by the context manager
        self.cs__parent_id = threading.current_thread().ident
    except Exception:
        logger.exception("Failed to instrument thread start")

    return orig_func(self, *args, **kwargs)


def _bootstrap_inner(orig_func, patch_policy, self, *args, **kwargs):
    # The new thread inherits the scope from the thread that created it
    try:
        cs_str.set_thread_scope(self.cs__parent_scope)
    except Exception:
        logger.exception("Failed to set thread scope")

    result = orig_func(self, *args, **kwargs)

    try:
        cs_str.destroy_thread_scope()
    except Exception:
        logger.exception("Failed to tear down thread scope")

    # We expect result to be None, but this is done for consistency/safety
    return result


def patch_threading(threading_module):
    patch_cls_or_instance(threading_module.Thread, START_METHOD, start)
    # This instruments the method that actually runs inside the system thread
    patch_cls_or_instance(threading_module.Thread, BOOTSTRAP_METHOD, _bootstrap_inner)


def register_patches():
    register_post_import_hook(patch_threading, "threading")


def reverse_patches():
    threading = sys.modules.get("threading")
    if not threading:
        return

    patch_manager.reverse_patches_by_owner(threading.Thread)



# -*- coding: utf-8 -*-
# Copyright © 2021 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import functools
import sys
from contrast.extern.wrapt import register_post_import_hook

from contrast.agent.policy import patch_manager
from contrast.utils.patch_utils import patch_cls_or_instance
from contrast.utils.assess.stream_utils import stream_patch_body

from contrast.extern import structlog as logging

logger = logging.getLogger("contrast")


def patched_func(orig_func, patch_policy, self, *args, **kwargs):
    bound_method = functools.partial(orig_func, self)
    return stream_patch_body(orig_func.__name__, bound_method, self, *args, **kwargs)


def patch_method(io_type, method_name):
    patch_cls_or_instance(io_type, method_name, patched_func)


def patch_getvalue(io_module):
    patch_method(io_module.StringIO, "getvalue")
    patch_method(io_module.BytesIO, "getvalue")


def patch_io(io_module):
    """
    Apply patches to methods of builtin stream types
    """
    for io_type in [io_module.StringIO, io_module.BytesIO]:
        patch_method(io_type, "write")

    # No need to hook StringIO.writelines because it is implemented as str.join under
    # the hood, so we already propagate for free. Unfortunately this might make the
    # reporting look a little odd, so we maybe should consider another solution later.
    patch_method(io_module.BytesIO, "writelines")

    # This patch exists solely for the purposes of working around our inability
    # to patch getvalue in StringIO and BytesIO using funchook.
    patch_getvalue(io_module)


def register_patches():
    register_post_import_hook(patch_io, "io")


def reverse_patches():
    io_module = sys.modules.get("io")
    if not io_module:
        return

    patch_manager.reverse_patches_by_owner(io_module)
    patch_manager.reverse_patches_by_owner(io_module.BytesIO)
    patch_manager.reverse_patches_by_owner(io_module.StringIO)



# -*- coding: utf-8 -*-
# Copyright © 2021 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.wsgi.middleware import WSGIMiddleware
from contrast.agent.middlewares.route_coverage.falcon_routes import (
    create_falcon_routes,
    get_view_method,
    build_falcon_route,
)
from contrast.utils.decorators import fail_safely, fail_quietly

from contrast.extern import structlog as logging

logger = logging.getLogger("contrast")


class FalconMiddleware(WSGIMiddleware):
    def __init__(self, app):
        # Since Falcon is WSGI-based, there is no way to retrieve the app name.
        # Use common config to define an app name.
        super(FalconMiddleware, self).__init__(app, app_name="Falcon Application")

        self.falcon_app = app
        # used for route coverage only for falcon middleware
        self.endpoint_cls = None

    @fail_safely("Unable to get route coverage", return_value={})
    def get_route_coverage(self):
        return create_falcon_routes(self.falcon_app)

    @fail_quietly("Unable to get Falcon view func")
    def get_view_func(self, request):
        path = request.path
        if not path:
            return None

        route_info = self.falcon_app._router.find(path)
        if not route_info:
            return None

        self.endpoint_cls, _, _, _ = route_info
        view_func = get_view_method(self.endpoint_cls, request.method)
        return view_func

    @fail_safely("Unable to build route", return_value="")
    def build_route(self, view_func, url):
        return build_falcon_route(view_func, self.endpoint_cls)



# -*- coding: utf-8 -*-
# Copyright © 2021 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import cgi

from contrast.extern.six import integer_types, iteritems, PY2
from contrast.extern.six.moves.collections_abc import Mapping

import contrast
from contrast.agent.assess.adjusted_span import AdjustedSpan
from contrast.agent.assess.policy.source_node import SourceNode
from contrast.agent.assess.properties import Properties
from contrast.agent.assess.utils import (
    get_properties,
    get_last_event_ids_from_sources,
    is_trackable,
    set_properties,
    track_string,
)
from contrast.agent.policy.constants import OBJECT, RETURN
from contrast.agent.settings_state import SettingsState
from contrast.utils.assess.duck_utils import is_iterable, safe_getattr, safe_iterator
from contrast.utils.decorators import fail_safely
from contrast.utils.string_utils import truncated_signature
from tempfile import SpooledTemporaryFile

from contrast.extern import structlog as logging

logger = logging.getLogger("contrast")

try:
    from werkzeug.datastructures import EnvironHeaders, FileStorage

    EnvironHeaders_types = (EnvironHeaders,)
    FileStorage_types = (FileStorage,)
except ImportError:
    EnvironHeaders_types = ()
    FileStorage_types = ()


def apply(
    source_nodes, self_obj, ret, args, kwargs=None
):  # pylint: disable=redefined-builtin
    """
    Apply node into all sources to track all new strings coming in from a request
    """
    context = contrast.CS__CONTEXT_TRACKER.current()
    if context is None or not SettingsState().is_assess_enabled():
        return

    for node in source_nodes:

        # args[0] is `self` for instance methods
        args = args[1:] if node.instance_method else args

        target = node.get_matching_first_target(self_obj, ret, args, kwargs)

        source_type = node.type
        if isinstance(target, FileStorage_types):
            # We don't try to propagate through werkzeug's file parsing logic, because
            # it causes massive slowdown. Instead, if we see a FileStorage object
            # come out of a multidict, we track it. We need to track the underlying
            # stream, not the FileStorage object itself.
            target = target.stream if PY2 else target.stream._file
            source_type = "MULTIPART_CONTENT_DATA"

        if hasattr(target, "cs__source"):
            # Target is already a source, nothing to do
            if target.cs__source:
                return

            target.cs__source = True
            target.cs__source_type = source_type
            target.cs__source_tags = node.tags
            set_properties(target, Properties(target))

        cs__apply_source(
            context, node, target, self_obj, ret, args, kwargs, source_type=source_type
        )


def adjust_source_tags(tags, source_type, source_name):
    """
    Update source tags to account for Referer header

    The XSS spec dictates that, with a single exception, no HTTP headers (or keys)
    should trigger XSS due to a high rate of false positives. The one exception is the
    'Referer' header. In order to account for this, we add the CROSS_SITE tag if the
    given source appears to correspond to this header value.
    """
    is_header = source_type == "HEADER"
    # This accounts for various representations of referer header keys, including
    # 'Referer' and 'HTTP_REFERER'
    is_referer = is_header and source_name and source_name.lower().endswith("referer")

    # According to XSS spec, the only header that should have CROSS_SITE is Referer.
    return (
        list(tags) + ["CROSS_SITE"]
        if (is_referer and "CROSS_SITE" not in tags)
        else list(tags)
    )


@fail_safely("Error in apply_stream_source")
def apply_stream_source(method_name, target, self_obj, ret, args, kwargs):
    context = contrast.CS__CONTEXT_TRACKER.current()
    if context is None or not SettingsState().is_assess_enabled():
        return

    module = self_obj.__class__.__module__
    class_name = self_obj.__class__.__name__

    source_type = safe_getattr(self_obj, "cs__source_type", None) or "BODY"
    source_tags = safe_getattr(self_obj, "cs__source_tags", None) or ["UNTRUSTED"]

    node = SourceNode(
        module,
        class_name,
        True,
        method_name,
        "RETURN",
        source_type,
        None,
        tags=source_tags,
    )

    cs__apply_source(context, node, target, self_obj, ret, args, kwargs)


NAMED_SOURCE_TYPES = [
    "PARAMETER",
    "PARAMETER_KEY",
    "HEADER",
    "HEADER_KEY",
    "COOKIE",
    "COOKIE_KEY",
    "MULTIPART_HEADER",
    "MULTIPART_HEADER_KEY",
    "MULTIPART_PARAMETER",
    "MULTIPART_PARAMETER_KEY",
]


def _get_source_name(node, self_obj, args):
    """
    Determine name for source

    Some sources require a source name since there can potentially be multiple values
    for that given source type (e.g. multiple headers, cookies, etc.).

    If no source name is explicitly given, but we have a dict/mapping object, then we
    try to use the dict/mapping key as the source name. This assumes that the
    instrumented dict/mapping method is something like __getitem__.

    Otherwise, use the method name (e.g. for `flask.wrappers.Request.full_path`, the
    source name will be "full_path").
    """
    if args and isinstance(self_obj, (dict, Mapping) + EnvironHeaders_types):
        return args[0]

    return node.method_name


def cs__apply_source(
    context,
    node,
    target,
    self_obj,
    ret,
    args,
    kwargs,
    source_name=None,
    source_type=None,
):
    # target: {param: test} / QueryDict
    # self_obj: QueryDict
    # ret: test
    # args: param

    logger.debug("In cs__apply_source for %r", node)
    if source_type is None:
        source_type = node.type

    if source_type in NAMED_SOURCE_TYPES and not source_name:
        source_name = _get_source_name(node, self_obj, args)

    _cs__apply_source(
        context,
        node,
        target,
        self_obj,
        ret,
        source_type,
        source_name,
        True,
        args,
        kwargs,
    )


@fail_safely("Unable to apply source for node")
def _cs__apply_source(
    context,
    node,
    target,
    self_obj,
    ret,
    source_type,
    source_name,
    use_source_key,
    args,
    kwargs,
):
    if not context or not node or target is None:
        return

    target_properties = get_properties(target)
    if target_properties is None and is_trackable(target):
        target_properties = track_string(target)
    if target_properties is not None:
        if not safe_getattr(target, "cs__source", False):
            for tag in adjust_source_tags(node.tags, source_type, source_name):
                length = len(target)

                target_properties.add_tag(tag, AdjustedSpan(0, length))
                target_properties.add_properties(node.properties)

                logger.debug(
                    "Source %s detected: %s tagged with %s",
                    node.name,
                    truncated_signature(target),
                    str(tag),
                )

        parent_ids = get_parent_ids(node, self_obj, ret, args, kwargs)
        target_properties.build_event(
            node,
            target,
            self_obj,
            ret,
            args,
            kwargs,
            parent_ids,
            possible_key=None,
            source_type=source_type,
            source_name=source_name,
        )

        if is_trackable(target):
            set_properties(target, target_properties)

    elif isinstance(target, dict):
        # This logic is intended to handle the case where the source is a dict
        # that happens to represent request parameters, headers, or cookies.
        # In these cases we want the top-level keys of this dictionary to be
        # labeled with the appropriate data type (e.g. PARAMETER_KEY). This
        # behavior should not apply to recursive calls, however.
        source_key_type = key_type(source_type) if use_source_key else source_type
        for key, value in iteritems(target):
            _cs__apply_source(
                context,
                node,
                key,
                self_obj,
                ret,
                source_key_type,
                key,
                False,
                args,
                kwargs,
            )
            _cs__apply_source(
                context,
                node,
                value,
                self_obj,
                ret,
                source_type,
                key,
                False,
                args,
                kwargs,
            )

    elif isinstance(target, cgi.FieldStorage):
        # circular import
        from contrast.utils.assess.stream_utils import (  # noqa
            ContrastFileProxy,
            ContrastRawIOProxy,
            BaseStreamProxy,
        )

        if hasattr(target, "filename"):
            _cs__apply_source(
                context,
                node,
                target.filename,
                self_obj,
                ret,
                source_type,
                source_name,
                False,
                args,
                kwargs,
            )
        if hasattr(target, "file") and not isinstance(target.file, BaseStreamProxy):
            # In PY2, FieldStorage.file has type `file`. In PY3, it has type FileIO.
            if hasattr(target.file, "cs__source"):
                target.file.cs__source = True
            elif PY2:
                target.file = ContrastFileProxy(target.file)
            else:
                target.file = ContrastRawIOProxy(target.file)

            # TODO: PYT-921 fix so we get the proper source type for Pyramid files
            # target.file.cs__source_type = node.type
            target.file.cs__source_tags = node.tags

    elif isinstance(target, SpooledTemporaryFile):
        # circular import
        from contrast.utils.assess.stream_utils import ContrastFileProxy  # noqa

        target._file = ContrastFileProxy(target._file)
        target._file.cs__source_type = "MULTIPART_CONTENT_DATA"
        target._file.cs__source_tags = node.tags

    elif is_iterable(target):
        for value in safe_iterator(target):
            _cs__apply_source(
                context,
                node,
                value,
                self_obj,
                ret,
                source_type,
                source_name,
                False,
                args,
                kwargs,
            )


def value_of_source(source, self_obj, ret, args, kwargs):
    if source == OBJECT:
        return self_obj

    if source == RETURN:
        return ret

    if not args:
        return self_obj

    if args and isinstance(source, integer_types) and source < len(args):
        return args[source]

    if kwargs and source in kwargs:
        return kwargs[source]

    return None


def get_parent_ids(node, self_obj, ret, args, kwargs):
    if isinstance(node, SourceNode):
        return []

    sources = [
        value_of_source(source, self_obj, ret, args, kwargs) for source in node.sources
    ]
    return get_last_event_ids_from_sources(sources)


KEY_MAPPING = {
    "PARAMETER": "PARAMETER_KEY",
    "HEADER": "HEADER_KEY",
    "COOKIE": "COOKIE_KEY",
}

PARAMETER_TYPE = "PARAMETER"
PARAMETER_KEY_TYPE = "PARAMETER_KEY"
HEADER_TYPE = "HEADER"
HEADER_KEY_TYPE = "HEADER_KEY"
COOKIE_TYPE = "COOKIE"
COOKIE_KEY_TYPE = "COOKIE_KEY"
QUERYSTRING_TYPE = "QUERYSTRING"


def key_type(source_type):
    return KEY_MAPPING.get(source_type, source_type)



# -*- coding: utf-8 -*-
# Copyright © 2021 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.settings_state import SettingsState
from contrast.utils.digest_utils import Digest

from contrast.extern import structlog as logging

logger = logging.getLogger("contrast")


class BaseRule(object):
    """
    Base rule object that all assess rules will inherit
    """

    @property
    def name(self):
        return "assess-base-rule"

    @property
    def disabled(self):
        """
        Property indicating whether rule is disabled
        """
        return SettingsState().is_assess_rule_disabled(self.name)

    def add_events_to_finding(self, finding, events=None, **kwargs):
        if events is not None:
            finding.events.extend(events)

    def generate_preflight_hash(self, **kwargs):
        hasher = Digest()
        hasher.update(self.name)

        self.update_preflight_hash(hasher, **kwargs)

        return hasher.finish()

    def update_preflight_hash(self, hasher, **kwargs):
        """
        Update preflight hash with additional rule-specific data

        Child classes should override this method in order to customize the
        kind of data that is used to generate the preflight hash.

        @param hasher: Hash class to be updated with additional data
        @param **kwargs: Placeholder for keyword args used by child classes
        """
        raise NotImplementedError
